@@ -1,5 +1,5 @@
 /*
- * $Id: mod_ruby.c 37 2004-09-25 03:59:57Z shugo $
+ * $Id: mod_ruby.c 110 2006-02-18 00:35:03Z shugo $
  * Copyright (C) 2000  ZetaBITS, Inc.
  * Copyright (C) 2000  Information-technology Promotion Agency, Japan
  * Copyright (C) 2001  Shugo Maeda <shugo@modruby.net>
@@ -70,6 +70,7 @@ static char **my_environ;
 #undef environ
 #define environ my_environ
 #elif defined(__APPLE__)
+#include <crt_externs.h>
 #undef environ
 #define environ (*_NSGetEnviron())
 #define GET_ENVIRON(e) (e)
@@ -97,6 +98,8 @@ static VALUE orig_defout;
 RUBY_EXTERN VALUE rb_load_path;
 static VALUE default_load_path;
 
+RUBY_EXTERN VALUE rb_progname;
+
 static const char *default_kcode;
 
 static int ruby_is_running = 0;
@@ -145,55 +148,71 @@ static int ruby_header_parser_handler(request_rec *r);
 #endif
 static int ruby_post_read_request_handler(request_rec *r);
 
-static const command_rec ruby_cmds[] =
-{
-    {"RubyKanjiCode", ruby_cmd_kanji_code, NULL, OR_ALL, TAKE1,
-     "set $KCODE"},
-    {"RubyAddPath", ruby_cmd_add_path, NULL, OR_ALL, ITERATE,
-     "add path to $:"},
-    {"RubyRequire", ruby_cmd_require, NULL, OR_ALL, ITERATE,
-     "ruby script name, pulled in via require"},
-    {"RubyPassEnv", ruby_cmd_pass_env, NULL, RSRC_CONF, ITERATE,
-     "pass environment variables to ENV"},
-    {"RubySetEnv", ruby_cmd_set_env, NULL, OR_ALL, TAKE2,
-     "Ruby ENV key and value" },
-    {"RubyTimeOut", ruby_cmd_timeout, NULL, RSRC_CONF, TAKE1,
-     "time to wait execution of ruby script"},
-    {"RubySafeLevel", ruby_cmd_safe_level, NULL, OR_ALL, TAKE1,
-     "set default $SAFE"},
-    {"RubyOutputMode", ruby_cmd_output_mode, NULL, OR_ALL, TAKE1,
-     "set output mode (nosync|sync|syncheader)"},
-    {"RubyHandler", ruby_cmd_handler, NULL, OR_ALL, TAKE1,
-     "set ruby handler object"},
-    {"RubyTransHandler", ruby_cmd_trans_handler, NULL, RSRC_CONF, TAKE1,
-     "set translation handler object"},
-    {"RubyAuthenHandler", ruby_cmd_authen_handler, NULL, OR_ALL, TAKE1,
-     "set authentication handler object"},
-    {"RubyAuthzHandler", ruby_cmd_authz_handler, NULL, OR_ALL, TAKE1,
-     "set authorization handler object"},
-    {"RubyAccessHandler", ruby_cmd_access_handler, NULL, OR_ALL, TAKE1,
-     "set access checker object"},
-    {"RubyTypeHandler", ruby_cmd_type_handler, NULL, OR_ALL, TAKE1,
-     "set type checker object"},
-    {"RubyFixupHandler", ruby_cmd_fixup_handler, NULL, OR_ALL, TAKE1,
-     "set fixup handler object"},
-    {"RubyLogHandler", ruby_cmd_log_handler, NULL, OR_ALL, TAKE1,
-     "set log handler object"},
-    {"RubyHeaderParserHandler", ruby_cmd_header_parser_handler,
-     NULL, OR_ALL, TAKE1,
-     "set header parser object"},
-    {"RubyPostReadRequestHandler", ruby_cmd_post_read_request_handler,
-     NULL, RSRC_CONF, TAKE1,
-     "set post-read-request handler object"},
-    {"RubyInitHandler", ruby_cmd_init_handler,
-     NULL, OR_ALL, TAKE1,
-     "set init handler object"},
-    {"RubyCleanupHandler", ruby_cmd_cleanup_handler,
-     NULL, OR_ALL, TAKE1,
-     "set cleanup handler object"},
-    {"RubyChildInitHandler", ruby_cmd_child_init_handler,
-     NULL, RSRC_CONF, TAKE1,
-     "set child init handler object"},
+#ifndef AP_INIT_TAKE1
+# define AP_INIT_TAKE1(directive, func, mconfig, where, help) \
+    { directive, func, mconfig, where, TAKE1, help }
+# define AP_INIT_TAKE2(directive, func, mconfig, where, help) \
+    { directive, func, mconfig, where, TAKE2, help }
+# define AP_INIT_ITERATE(directive, func, mconfig, where, help) \
+    { directive, func, mconfig, where, ITERATE, help }
+# define AP_INIT_FLAG(directive, func, mconfig, where, help) \
+    { directive, func, mconfig, where, FLAG, help }
+#endif
+
+static const command_rec ruby_cmds[] = {
+    AP_INIT_TAKE1("RubyKanjiCode", (const char *(*)())ruby_cmd_kanji_code, NULL, OR_ALL,
+     "set $KCODE"),
+    AP_INIT_ITERATE("RubyAddPath", ruby_cmd_add_path, NULL, OR_ALL,
+     "add path to $:"),
+    AP_INIT_ITERATE("RubyRequire", ruby_cmd_require, NULL, OR_ALL,
+     "ruby script name, pulled in via require"),
+    AP_INIT_ITERATE("RubyPassEnv", ruby_cmd_pass_env, NULL, RSRC_CONF,
+     "pass environment variables to ENV"),
+    AP_INIT_TAKE2("RubySetEnv", ruby_cmd_set_env, NULL, OR_ALL,
+     "Ruby ENV key and value" ),
+    AP_INIT_TAKE1("RubyTimeOut", ruby_cmd_timeout, NULL, RSRC_CONF,
+     "time to wait execution of ruby script"),
+    AP_INIT_TAKE1("RubySafeLevel", ruby_cmd_safe_level, NULL, OR_ALL,
+     "set default $SAFE"),
+    AP_INIT_TAKE1("RubyOutputMode", ruby_cmd_output_mode, NULL, OR_ALL,
+     "set output mode (nosync|sync|syncheader)"),
+    AP_INIT_FLAG("RubyRestrictDirectives", ruby_cmd_restrict_directives, NULL, 
+     RSRC_CONF, "whether Ruby* directives are restricted from .htaccess"),
+    AP_INIT_TAKE2("RubyOption", ruby_cmd_option, NULL, OR_ALL,
+     "set option for application"),
+    AP_INIT_FLAG("RubyGcPerRequest", ruby_cmd_gc_per_request, NULL, 
+     OR_ALL, "whether to call GC for each request"),
+    AP_INIT_TAKE1("RubyHandler", ruby_cmd_handler, NULL, OR_ALL,
+     "set ruby handler object"),
+    AP_INIT_TAKE1("RubyTransHandler", ruby_cmd_trans_handler, NULL, OR_ALL,
+     "set translation handler object"),
+    AP_INIT_TAKE1("RubyAuthenHandler", ruby_cmd_authen_handler, NULL, OR_ALL,
+     "set authentication handler object"),
+    AP_INIT_TAKE1("RubyAuthzHandler", ruby_cmd_authz_handler, NULL, OR_ALL,
+     "set authorization handler object"),
+    AP_INIT_TAKE1("RubyAccessHandler", ruby_cmd_access_handler, NULL, OR_ALL,
+     "set access checker object"),
+    AP_INIT_TAKE1("RubyTypeHandler", ruby_cmd_type_handler, NULL, OR_ALL,
+     "set type checker object"),
+    AP_INIT_TAKE1("RubyFixupHandler", ruby_cmd_fixup_handler, NULL, OR_ALL,
+     "set fixup handler object"),
+    AP_INIT_TAKE1("RubyLogHandler", ruby_cmd_log_handler, NULL, OR_ALL,
+     "set log handler object"),
+    AP_INIT_TAKE1("RubyHeaderParserHandler", ruby_cmd_header_parser_handler,
+     NULL, OR_ALL,
+     "set header parser object"),
+    AP_INIT_TAKE1("RubyPostReadRequestHandler", ruby_cmd_post_read_request_handler,
+     NULL, RSRC_CONF,
+     "set post-read-request handler object"),
+    AP_INIT_TAKE1("RubyInitHandler", ruby_cmd_init_handler,
+     NULL, OR_ALL,
+     "set init handler object"),
+    AP_INIT_TAKE1("RubyCleanupHandler", ruby_cmd_cleanup_handler,
+     NULL, OR_ALL,
+     "set cleanup handler object"),
+    AP_INIT_TAKE1("RubyChildInitHandler", ruby_cmd_child_init_handler,
+     NULL, RSRC_CONF,
+     "set child init handler object"),
     {NULL}
 };
 
@@ -320,12 +339,12 @@ VALUE rb_protect_funcall(VALUE recv, ID mid, int *state, int argc, ...)
 static void get_error_pos(VALUE str)
 {
     char buff[BUFSIZ];
-    ID last_func = rb_frame_last_func();
+    ID this_func = rb_frame_this_func();
 
     if (ruby_sourcefile) {
-	if (last_func) {
+	if (this_func) {
 	    snprintf(buff, BUFSIZ, "%s:%d:in `%s'", ruby_sourcefile, ruby_sourceline,
-		     rb_id2name(last_func));
+		     rb_id2name(this_func));
 	}
 	else {
 	    snprintf(buff, BUFSIZ, "%s:%d", ruby_sourcefile, ruby_sourceline);
@@ -501,13 +520,15 @@ void ruby_log_error_string(server_rec *s, VALUE errmsg)
 
 static void handle_error(request_rec *r, int state)
 {
-    VALUE errmsg, reqobj;
+    ruby_request_config *rconf;
+    VALUE errmsg;
 
     errmsg = ruby_get_error_info(state);
     if (r->request_config) {
-	reqobj = (VALUE) ap_get_module_config(r->request_config, &ruby_module);
-	if (reqobj)
-	    rb_apache_request_set_error(reqobj, errmsg, ruby_errinfo);
+	rconf = get_request_config(r);
+	if (!NIL_P(rconf->request_object))
+	    rb_apache_request_set_error(rconf->request_object,
+					errmsg, ruby_errinfo);
     }
     ruby_log_error_string(r->server, errmsg);
 }
@@ -543,7 +564,7 @@ void mod_ruby_setup_loadpath(ruby_server_config *sconf,
     }
 }
 
-static int ruby_require_directly(char *filename,
+static int ruby_require_directly(const char *filename,
 				 ruby_server_config *sconf,
 				 ruby_dir_config *dconf)
 {
@@ -695,7 +716,9 @@ static void ruby_init_interpreter(server_rec *s)
 
 static void dso_unload(void *handle)
 {
-#if defined(HPUX) || defined(HPUX10) || defined(HPUX11)
+#if defined(_WIN32)
+    FreeLibrary((HANDLE) handle);
+#elif defined(HPUX) || defined(HPUX10) || defined(HPUX11)
     shl_unload((shl_t) handle);
 #elif defined(HAVE_DYLD)
     NSUnLinkModule(handle, FALSE);
@@ -818,7 +841,7 @@ static APR_CLEANUP_RETURN_TYPE ruby_child_cleanup(void *data)
 
 static request_rec *fake_request_rec(server_rec *s, pool *p, char *hook)
 {
-    request_rec *r = (request_rec *) ap_pcalloc(p, sizeof(request_rec));
+    request_rec *r = (request_rec *) apr_pcalloc(p, sizeof(request_rec));
     r->pool = p; 
     r->server = s;
     r->per_dir_config = NULL;
@@ -888,7 +911,7 @@ static void ruby_child_init(server_rec *s, pool *p)
 #if APR_HAS_THREADS
 	}
 #endif
-	ap_register_cleanup(p, NULL, ruby_child_cleanup, ap_null_cleanup);
+	apr_pool_cleanup_register(p, NULL, ruby_child_cleanup, apr_pool_cleanup_null);
     }
 
     r = fake_request_rec(s, p, "RubyChildInitHandler");
@@ -900,14 +923,14 @@ static void ruby_child_init(server_rec *s, pool *p)
 static void mod_ruby_clearenv(pool *p)
 {
     char **env;
-    array_header *names = ap_make_array(p, 1, sizeof(char*));
+    array_header *names = apr_array_make(p, 1, sizeof(char*));
     int i;
     
     env = GET_ENVIRON(environ);
     while (*env) {
         char *s = strchr(*env, '=');
         if (s) {
-            *(char **) ap_push_array(names) = ap_pstrndup(p, *env, s - *env);
+            *(char **) apr_array_push(names) = apr_pstrndup(p, *env, s - *env);
         }
         env++;
     }
@@ -937,7 +960,7 @@ static void setenv_from_table(table *tbl)
     table_entry *env;
     int i;
 
-    env_arr = ap_table_elts(tbl);
+    env_arr = apr_table_elts(tbl);
     env = (table_entry *) env_arr->elts;
     for (i = 0; i < env_arr->nelts; i++) {
 	if (env[i].key == NULL)
@@ -1001,19 +1024,22 @@ typedef struct run_safely_arg {
     void *arg;
 } run_safely_arg_t;
 
-static VALUE run_safely_0(void *arg)
+static VALUE run_safely_0(VALUE arg)
 {
     run_safely_arg_t *rsarg = (run_safely_arg_t *) arg;
     struct timeout_arg targ;
-    VALUE timeout_thread;
+    VALUE timeout_thread = Qnil;
     VALUE result;
 
     rb_set_safe_level(rsarg->safe_level);
-    targ.thread = rb_thread_current();
-    targ.timeout = rsarg->timeout;
-    timeout_thread = rb_thread_create(do_timeout, (void *) &targ);
+    if (rsarg->timeout > 0) {
+	targ.thread = rb_thread_current();
+	targ.timeout = rsarg->timeout;
+	timeout_thread = rb_thread_create(do_timeout, (void *) &targ);
+    }
     result = (*rsarg->func)(rsarg->arg);
-    rb_protect_funcall(timeout_thread, rb_intern("kill"), NULL, 0);
+    if (!NIL_P(timeout_thread))
+	rb_protect_funcall(timeout_thread, rb_intern("kill"), NULL, 0);
     return result;
 }
 
@@ -1031,8 +1057,13 @@ static int run_safely(int safe_level, int timeout,
 #if defined(HAVE_SETITIMER)
     rb_thread_start_timer();
 #endif
-    thread = rb_thread_create(run_safely_0, &rsarg);
-    ret = rb_protect_funcall(thread, rb_intern("value"), &state, 0);
+    if (safe_level > ruby_safe_level) {
+	thread = rb_thread_create(run_safely_0, &rsarg);
+	ret = rb_protect_funcall(thread, rb_intern("value"), &state, 0);
+    }
+    else {
+	ret = rb_protect(run_safely_0, (VALUE) &rsarg, &state);
+    }
     rb_protect(kill_threads, Qnil, NULL);
 #if defined(HAVE_SETITIMER)
     rb_thread_stop_timer();
@@ -1042,11 +1073,53 @@ static int run_safely(int safe_level, int timeout,
     return state;
 }
 
+static table *save_env(pool *p)
+{
+    char **env, *k, *v;
+    table *result;
+    
+    env = GET_ENVIRON(environ);
+    result = apr_table_make(p, 1);
+    while (*env) {
+        char *s = strchr(*env, '=');
+        if (s) {
+            k = apr_pstrndup(p, *env, s - *env);
+            v = apr_pstrdup(p, s + 1);
+	    apr_table_set(result, k, v);
+        }
+        env++;
+    }
+    return result;
+}
+
+static void restore_env(pool *p, table *env)
+{
+    const array_header *hdrs_arr;
+    table_entry *hdrs;
+    int i;
+
+    mod_ruby_clearenv(p);
+    hdrs_arr = apr_table_elts(env);
+    hdrs = (table_entry *) hdrs_arr->elts;
+    for (i = 0; i < hdrs_arr->nelts; i++) {
+	if (hdrs[i].key == NULL)
+	    continue;
+	mod_ruby_setenv(hdrs[i].key, hdrs[i].val);
+    }
+}
+
 static void per_request_init(request_rec *r)
 {
+    ruby_request_config *rconf;
     ruby_server_config *sconf;
     ruby_dir_config *dconf;
 
+    if (r->request_config) {
+	rconf = apr_palloc(r->pool, sizeof(ruby_request_config));
+	rconf->saved_env = save_env(r->pool);
+	rconf->request_object = Qnil;
+	ap_set_module_config(r->request_config, &ruby_module, rconf);
+    }
     dconf = get_dir_config(r);
     sconf = get_server_config(r->server);
     mod_ruby_setup_loadpath(sconf, dconf);
@@ -1059,6 +1132,8 @@ static void per_request_init(request_rec *r)
 #if RUBY_VERSION_CODE < 180
     rb_defout = rb_request;
 #endif
+    if (r->filename)
+	rb_progname = rb_tainted_str_new2(r->filename);
 }
 
 static VALUE exec_end_proc(VALUE arg)
@@ -1069,6 +1144,7 @@ static VALUE exec_end_proc(VALUE arg)
 
 static void per_request_cleanup(request_rec *r, int flush)
 {
+    ruby_dir_config *dconf = get_dir_config(r);
     VALUE reqobj;
 
     while (r->next)
@@ -1095,7 +1171,13 @@ static void per_request_cleanup(request_rec *r, int flush)
 #endif
     }
     rb_set_kcode(default_kcode);
-    rb_gc();
+    if (r->request_config) {
+	ruby_request_config *rconf = get_request_config(r);
+	restore_env(r->pool, rconf->saved_env);
+    }
+    rb_progname = Qnil;
+    if (dconf->gc_per_request)
+	rb_gc();
 }
 
 typedef struct handler_0_arg {
@@ -1160,6 +1242,10 @@ static void *ruby_handler_internal(handler_internal_arg_t *iarg)
     handler_0_arg_t arg;
     int i, handlers_len;
     char **handlers;
+    int timeout;
+#ifndef APACHE2
+    static int warned_timeout = 0;
+#endif
 
     sconf = get_server_config(r->server);
     dconf = get_dir_config(r);
@@ -1174,7 +1260,18 @@ static void *ruby_handler_internal(handler_internal_arg_t *iarg)
 	arg.handler = handlers[i];
 	arg.mid = mid;
 	ap_soft_timeout("call ruby handler", r);
-	if ((state = run_safely(safe_level, sconf->timeout,
+	timeout = sconf->timeout;
+#ifndef APACHE2
+	if (timeout >= r->server->timeout) {
+	    if (!warned_timeout)
+		ruby_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO,
+			       r->server,
+			       "disabled RubyTimeOut: RubyTimeOut >= TimeOut");
+	    timeout = 0;
+	    warned_timeout = 1;
+	}
+#endif
+	if ((state = run_safely(safe_level, timeout,
 				ruby_handler_0, &arg, &ret)) == 0) {
 	    iarg->retval = NUM2INT(ret);
 	}
@@ -1199,7 +1296,7 @@ static int ruby_handler(request_rec *r,
     if (handlers_arr == NULL)
 	return DECLINED;
 
-    arg = ap_palloc(r->pool, sizeof(handler_internal_arg_t));
+    arg = apr_palloc(r->pool, sizeof(handler_internal_arg_t));
     arg->r = r;
     arg->handlers_arr = handlers_arr;
     arg->mid = mid;
@@ -1346,13 +1443,13 @@ static int ruby_post_read_request_handler(request_rec *r)
     ruby_dir_config *dconf = get_dir_config(r);
     int retval;
 
-    ap_register_cleanup(r->pool, (void *) r, ruby_cleanup_handler, 
-			ap_null_cleanup);
+    apr_pool_cleanup_register(r->pool, (void *) r, ruby_cleanup_handler, 
+			apr_pool_cleanup_null);
 
     if (dconf->ruby_init_handler) {
 	retval = ruby_handler(r, dconf->ruby_init_handler,
 			      rb_intern("init"), 1, 0);
-	ap_table_set(r->notes, "ruby_init_ran", "true");
+	apr_table_set(r->notes, "ruby_init_ran", "true");
 	if (retval != OK && retval != DECLINED)
 	    return retval;
     }
