@@ -1,5 +1,5 @@
 /*
- * $Id: mod_ruby.h 38 2004-09-26 16:14:29Z shugo $
+ * $Id: mod_ruby.h 107 2006-02-16 14:57:24Z shugo $
  * Copyright (C) 2000  ZetaBITS, Inc.
  * Copyright (C) 2000  Information-technology Promotion Agency, Japan
  * Copyright (C) 2000  Shugo Maeda <shugo@modruby.net>
@@ -30,6 +30,22 @@
 #ifndef MOD_RUBY_H
 #define MOD_RUBY_H
 
+#include "ruby.h"
+#include "rubyio.h"
+#include "version.h"
+#include "util.h"
+#include "intern.h"
+
+#if RUBY_VERSION_CODE < 190 || RUBY_RELEASE_CODE <= 20050304
+#define rb_frame_this_func() rb_frame_last_func()
+#endif
+
+#if defined(_WIN32)
+#undef pid_t
+#undef uid_t
+#undef gid_t
+#endif
+
 #include "httpd.h"
 #include "http_config.h"
 #include "http_core.h"
@@ -39,12 +55,6 @@
 #include "http_request.h"
 #include "util_script.h"
 
-#include "ruby.h"
-#include "rubyio.h"
-#include "version.h"
-#include "util.h"
-#include "intern.h"
-
 #ifndef RUBY_EXTERN
 #define RUBY_EXTERN EXTERN
 #endif
@@ -89,6 +99,34 @@ typedef apr_table_entry_t table_entry;
 #define APR_CLEANUP_RETURN_SUCCESS() return
 
 #include "http_conf_globals.h"
+#define apr_palloc(p, size) ap_palloc(p, size)
+#define apr_pcalloc(p, type) ap_palloc(p, type)
+#define apr_pool_cleanup_register ap_register_cleanup
+#define apr_pool_cleanup_null ap_null_cleanup
+#define apr_array_make(p, n, size) ap_make_array(p, n, size)
+#define apr_array_push(ary) ap_push_array(ary)
+#define apr_array_append(p, ary1, ary2) ap_append_arrays(p, ary1, ary2)
+#define apr_table_make(p, n) ap_make_table(p, n)
+#define apr_table_elts(tbl) ap_table_elts(tbl)
+#define apr_table_get(tbl, key) ap_table_get(tbl, key)
+#define apr_table_add(tbl, key, val) ap_table_add(tbl, key, val)
+#define apr_table_set(tbl, key, val) ap_table_set(tbl, key, val)
+#define apr_table_setn(tbl, key, val) ap_table_setn(tbl, key, val)
+#define apr_table_unset(tbl, key) ap_table_unset(tbl, key)
+#define apr_table_overlay(p, tbl1, tbl2) ap_overlay_tables(p, tbl1, tbl2)
+#define apr_table_merge(p, tbl1, tbl2) ap_table_merge(p, tbl1, tbl2)
+#define apr_table_do ap_table_do
+#define apr_table_clear(tbl) ap_clear_table(tbl)
+#define apr_pstrdup(p, s) ap_pstrdup(p, s)
+#define apr_pstrndup(p, s, n) ap_pstrndup(p, s, n)
+#define apr_pstrcat ap_pstrcat
+#define apr_psprintf ap_psprintf
+#define apr_isspace(c) ap_isspace(c)
+#define apr_isdigit(c) ap_isdigit(c)
+#define apr_isxdigit(c) ap_isxdigit(c)
+#define apr_day_snames ap_day_snames
+#define apr_month_snames ap_month_snames
+#define apr_time_sec(t) (t)
 
 #endif
 
@@ -125,7 +163,7 @@ typedef apr_table_entry_t table_entry;
 #include "apache_multipart_buffer.h"
 #include "apache_cookie.h"
 
-#define MOD_RUBY_STRING_VERSION "mod_ruby/1.2.3"
+#define MOD_RUBY_STRING_VERSION "mod_ruby/1.2.5"
 #define RUBY_GATEWAY_INTERFACE "CGI-Ruby/1.1"
 
 typedef struct {
@@ -133,6 +171,7 @@ typedef struct {
     table *env;
     int timeout;
     array_header *ruby_child_init_handler;
+    int restrict_directives;
 } ruby_server_config;
 
 typedef struct {
@@ -141,6 +180,8 @@ typedef struct {
     int safe_level;
     int output_mode;
     array_header *load_path;
+    table *options;
+    int gc_per_request;
     array_header *ruby_handler;
     array_header *ruby_trans_handler;
     array_header *ruby_authen_handler;
@@ -156,13 +197,20 @@ typedef struct {
 } ruby_dir_config;
 
 typedef struct {
-    char *filename;
+    const char *filename;
     ruby_server_config *server_config;
     ruby_dir_config *dir_config;
 } ruby_library_context;
 
-#define MR_DEFAULT_TIMEOUT 270
+typedef struct {
+    table *saved_env;
+    VALUE request_object;
+} ruby_request_config;
+
+#define MR_DEFAULT_TIMEOUT 0
 #define MR_DEFAULT_SAFE_LEVEL 1
+#define MR_DEFAULT_RESTRICT_DIRECTIVES 0
+#define MR_DEFAULT_GC_PER_REQUEST 0
 
 #define MR_OUTPUT_DEFAULT	0
 #define MR_OUTPUT_NOSYNC	1
@@ -201,6 +249,9 @@ void mod_ruby_setup_loadpath(ruby_server_config *sconf,
 	   ((ruby_dir_config *) ap_get_module_config(r->per_dir_config, \
 						    &ruby_module)) : \
 	   NULL)
+#define get_request_config(r) \
+	((ruby_request_config *) ap_get_module_config(r->request_config, \
+						     &ruby_module))
 
 #if APR_HAS_THREADS
 typedef void *(*ruby_interp_func_t)(void*);
