@@ -1,5 +1,5 @@
 /*
- * $Id: request.c,v 1.46 2004/01/27 13:53:33 shugo Exp $
+ * $Id: request.c 104 2006-01-31 17:52:04Z shugo $
  * Copyright (C) 2000  ZetaBITS, Inc.
  * Copyright (C) 2000  Information-technology Promotion Agency, Japan
  * Copyright (C) 2001-2003  Shugo Maeda <shugo@modruby.net>
@@ -41,11 +41,9 @@
 static VALUE rb_eApachePrematureChunkEndError;
 VALUE rb_cApacheRequest;
 
-#ifdef HAVE_LIBAPREQ
 /* Parse options */
 static ID id_post_max, id_disable_uploads,
     id_temp_dir, id_hook_data, id_upload_hook;
-#endif
 
 typedef struct request_data {
     request_rec *request;
@@ -61,14 +59,13 @@ typedef struct request_data {
     VALUE attributes;
     VALUE error_message;
     VALUE exception;
-#ifdef HAVE_LIBAPREQ
     ApacheRequest *apreq;
     VALUE upload_hook;
     VALUE upload_hook_arg;
     VALUE upload_table;
     VALUE cookies;
     VALUE param_table;
-#endif
+    VALUE options;
 } request_data;
 
 #define REQ_SYNC_HEADER     FL_USER1
@@ -79,6 +76,19 @@ typedef struct request_data {
 #define REQUEST_STRING_ATTR_READER(fname, member) \
 	DEFINE_STRING_ATTR_READER(fname, request_data, request->member)
 
+#ifdef APACHE2
+#define REQUEST_STRING_ATTR_WRITER(fname, member) \
+static VALUE fname(VALUE self, VALUE val) \
+{ \
+    request_data *data; \
+    Check_Type(val, T_STRING); \
+    data = get_request_data(self); \
+    data->request->member = \
+	apr_pstrndup(data->request->pool, \
+		    RSTRING(val)->ptr, RSTRING(val)->len); \
+    return val; \
+}
+#else
 #define REQUEST_STRING_ATTR_WRITER(fname, member) \
 static VALUE fname(VALUE self, VALUE val) \
 { \
@@ -90,6 +100,7 @@ static VALUE fname(VALUE self, VALUE val) \
 		    RSTRING(val)->ptr, RSTRING(val)->len); \
     return val; \
 }
+#endif
 
 #define REQUEST_INT_ATTR_READER(fname, member) \
 	DEFINE_INT_ATTR_READER(fname, request_data, request->member)
@@ -111,33 +122,41 @@ static void request_mark(request_data *data)
     rb_gc_mark(data->attributes);
     rb_gc_mark(data->error_message);
     rb_gc_mark(data->exception);
-#ifdef HAVE_LIBAPREQ
     rb_gc_mark(data->upload_hook);
     rb_gc_mark(data->upload_hook_arg);
     rb_gc_mark(data->upload_table);
     rb_gc_mark(data->cookies);
     rb_gc_mark(data->param_table);
-#endif
+    rb_gc_mark(data->options);
 }
 
 static APR_CLEANUP_RETURN_TYPE cleanup_request_object(void *data)
 {
     request_rec *r = (request_rec *) data;
+    ruby_request_config *rconf;
     VALUE reqobj;
 
-    reqobj = (VALUE) ap_get_module_config(r->request_config, &ruby_module);
-    if (reqobj == 0) APR_CLEANUP_RETURN_SUCCESS();
+    if (r->request_config == NULL)
+	APR_CLEANUP_RETURN_SUCCESS();
+    rconf = get_request_config(r);
+    if (rconf == NULL) APR_CLEANUP_RETURN_SUCCESS();
+    reqobj = rconf->request_object;
     if (TYPE(reqobj) == T_DATA) {
 	free(RDATA(reqobj)->data);
 	RDATA(reqobj)->data = NULL;
     }
-    ap_set_module_config(r->request_config, &ruby_module, 0);
+    ap_set_module_config(r->request_config, &ruby_module, NULL);
 #if APR_HAS_THREADS
-    ruby_call_interpreter(r->pool,
-			  (ruby_interp_func_t) rb_apache_unregister_object,
-			  (void *) reqobj, NULL, 0);
-#else
-    rb_apache_unregister_object(reqobj);
+    if (ruby_is_threaded_mpm) {
+	ruby_call_interpreter(r->pool,
+			      (ruby_interp_func_t) rb_apache_unregister_object,
+			      (void *) reqobj, NULL, 0);
+    }
+    else {
+#endif
+	rb_apache_unregister_object(reqobj);
+#if APR_HAS_THREADS
+    }
 #endif
     APR_CLEANUP_RETURN_SUCCESS();
 }
@@ -147,6 +166,9 @@ static VALUE apache_request_new(request_rec *r)
     ruby_dir_config *dconf = get_dir_config(r);
     request_data *data;
     VALUE obj;
+    const array_header *opts_arr;
+    table_entry *opts;
+    int i;
 
     obj = Data_Make_Struct(rb_cApacheRequest, request_data,
 			   request_mark, free, data);
@@ -163,46 +185,56 @@ static VALUE apache_request_new(request_rec *r)
     data->attributes = Qnil;
     data->error_message = Qnil;
     data->exception = Qnil;
-#ifdef HAVE_LIBAPREQ
     data->apreq = ApacheRequest_new( r );
     data->upload_hook = Qnil;
     data->upload_hook_arg = Qnil;
     data->upload_table = rb_hash_new();
     data->cookies = rb_hash_new();
     data->param_table = Qnil;
-#endif
+    data->options = rb_hash_new();
+    opts_arr = apr_table_elts(dconf->options);
+    opts = (table_entry *) opts_arr->elts;
+    for (i = 0; i < opts_arr->nelts; i++) {
+        if (opts[i].key == NULL)
+	    continue;
+	rb_hash_aset(data->options,
+		     rb_tainted_str_new2(opts[i].key),
+		     rb_tainted_str_new2(opts[i].val));
+    }
     
     rb_apache_register_object(obj);
-    ap_set_module_config(r->request_config, &ruby_module, (void *) obj);
-    ap_register_cleanup(r->pool, (void *) r,
-			cleanup_request_object, ap_null_cleanup);
-    switch (dconf->output_mode) {
-    case MR_OUTPUT_SYNC_HEADER:
-	FL_SET(obj, REQ_SYNC_HEADER);
-	break;
-    case MR_OUTPUT_SYNC:
-	FL_SET(obj, REQ_SYNC_HEADER);
-	FL_SET(obj, REQ_SYNC_OUTPUT);
-	break;
-    case MR_OUTPUT_NOSYNC:
-    default:
-	break;
+    if (r->request_config) {
+	ruby_request_config *rconf = get_request_config(r);
+	rconf->request_object = obj;
+    }
+    apr_pool_cleanup_register(r->pool, (void *) r,
+			cleanup_request_object, apr_pool_cleanup_null);
+    if (dconf) {
+	switch (dconf->output_mode) {
+	    case MR_OUTPUT_SYNC_HEADER:
+		FL_SET(obj, REQ_SYNC_HEADER);
+		break;
+	    case MR_OUTPUT_SYNC:
+		FL_SET(obj, REQ_SYNC_HEADER);
+		FL_SET(obj, REQ_SYNC_OUTPUT);
+		break;
+	    case MR_OUTPUT_NOSYNC:
+	    default:
+		break;
+	}
     }
     return obj;
 }
 
 VALUE rb_get_request_object(request_rec *r)
 {
-    VALUE reqobj;
-
     if (r == NULL) return Qnil;
-    reqobj = (VALUE) ap_get_module_config(r->request_config, &ruby_module);
-    if (reqobj) {
-	return reqobj;
-    }
-    else {
-	return apache_request_new(r);
+    if (r->request_config) {
+	ruby_request_config *rconf = get_request_config(r);
+	if (!NIL_P(rconf->request_object))
+	    return rconf->request_object;
     }
+    return apache_request_new(r);
 }
 
 static request_data *get_request_data(VALUE obj)
@@ -381,14 +413,22 @@ static VALUE request_printf(int argc, VALUE *argv, VALUE out)
 
 static VALUE request_puts _((int, VALUE*, VALUE));
 
+#if RUBY_VERSION_CODE >= 190 && RUBY_RELEASE_CODE > 20050304
+static VALUE request_puts_ary(VALUE ary, VALUE out, int recur)
+#else
 static VALUE request_puts_ary(VALUE ary, VALUE out)
+#endif
 {
     VALUE tmp;
     int i;
 
     for (i=0; i<RARRAY(ary)->len; i++) {
 	tmp = RARRAY(ary)->ptr[i];
+#if RUBY_VERSION_CODE >= 190 && RUBY_RELEASE_CODE > 20050304
+	if (recur) {
+#else
 	if (rb_inspecting_p(tmp)) {
+#endif
 	    tmp = STRING_LITERAL("[...]");
 	}
 	request_puts(1, &tmp, out);
@@ -412,7 +452,11 @@ static VALUE request_puts(int argc, VALUE *argv, VALUE out)
 	    line = STRING_LITERAL("nil");
 	    break;
 	  case T_ARRAY:
+#if RUBY_VERSION_CODE >= 190 && RUBY_RELEASE_CODE > 20050304
+	    rb_exec_recursive(request_puts_ary, argv[i], out);
+#else
 	    rb_protect_inspect(request_puts_ary, argv[i], out);
+#endif
 	    continue;
 	  default:
 	    line = argv[i];
@@ -574,11 +618,7 @@ static VALUE request_request_time(VALUE self)
     request_data *data;
 
     data = get_request_data(self);
-#ifdef APACHE2
     return rb_time_new(apr_time_sec(data->request->request_time), 0);
-#else
-    return rb_time_new(data->request->request_time, 0);
-#endif
 }
 
 static VALUE request_header_only(VALUE self)
@@ -596,7 +636,7 @@ static VALUE request_content_length(VALUE self)
 
     rb_warn("content_length is obsolete; use headers_in[\"Content-Length\"] instead");
     data = get_request_data(self);
-    s = ap_table_get(data->request->headers_in, "Content-Length");
+    s = apr_table_get(data->request->headers_in, "Content-Length");
     return s ? rb_cstr2inum(s, 10) : Qnil;
 }
 
@@ -612,7 +652,7 @@ static VALUE request_set_content_type(VALUE self, VALUE str)
 	Check_Type(str, T_STRING);
 	str = rb_funcall(str, rb_intern("downcase"), 0);
 	data->request->content_type =
-	    ap_pstrndup(data->request->pool,
+	    apr_pstrndup(data->request->pool,
 			RSTRING(str)->ptr, RSTRING(str)->len);
     }
     return str;
@@ -630,7 +670,7 @@ static VALUE request_set_content_encoding(VALUE self, VALUE str)
 	Check_Type(str, T_STRING);
 	str = rb_funcall(str, rb_intern("downcase"), 0);
 	data->request->content_encoding =
-	    ap_pstrndup(data->request->pool,
+	    apr_pstrndup(data->request->pool,
 			RSTRING(str)->ptr,
 			RSTRING(str)->len);
     }
@@ -665,12 +705,12 @@ static VALUE request_set_content_languages(VALUE self, VALUE ary)
 	    Check_Type(RARRAY(ary)->ptr[i], T_STRING);
 	}
 	data->request->content_languages =
-	    ap_make_array(data->request->pool, RARRAY(ary)->len, sizeof(char *));
+	    apr_array_make(data->request->pool, RARRAY(ary)->len, sizeof(char *));
 	for (i = 0; i < RARRAY(ary)->len; i++) {
 	    VALUE str = RARRAY(ary)->ptr[i];
 	    str = rb_funcall(str, rb_intern("downcase"), 0);
-	    *(char **) ap_push_array(data->request->content_languages) =
-		ap_pstrndup(data->request->pool,
+	    *(char **) apr_array_push(data->request->content_languages) =
+		apr_pstrndup(data->request->pool,
 			    RSTRING(str)->ptr,
 			    RSTRING(str)->len);
 	}
@@ -870,14 +910,25 @@ static VALUE request_should_client_block(VALUE self)
 static VALUE request_get_client_block(VALUE self, VALUE length)
 {
     request_data *data;
-    char *buf;
     int len;
+    VALUE result;
 
     data = get_request_data(self);
     len = NUM2INT(length);
-    buf = (char *) ap_palloc(data->request->pool, len);
-    len = ap_get_client_block(data->request, buf, len);
-    return rb_tainted_str_new(buf, len);
+    result = rb_str_buf_new(len);
+    len = ap_get_client_block(data->request, RSTRING(result)->ptr, len);
+    switch (len) {
+    case -1:
+	rb_raise(rb_eApachePrematureChunkEndError, "premature chunk end");
+	break;
+    case 0:
+	return Qnil;
+    default:
+	RSTRING(result)->ptr[len] = '\0';
+	RSTRING(result)->len = len;
+	OBJ_TAINT(result);
+	return result;
+    }
 }
 
 static VALUE read_client_block(request_rec *r, int len)
@@ -898,7 +949,7 @@ static VALUE read_client_block(request_rec *r, int len)
     if (ap_should_client_block(r)) {
 	if (len < 0)
 	    len = r->remaining;
-	buf = (char *) ap_palloc(r->pool, len);
+	buf = (char *) apr_palloc(r->pool, len);
 	result = rb_tainted_str_new("", 0);
 	while (len > 0) {
 	    nrd = ap_get_client_block(r, buf, len);
@@ -1152,7 +1203,7 @@ static VALUE request_soft_timeout(VALUE self, VALUE name)
 
     Check_Type(name, T_STRING);
     data = get_request_data(self);
-    s = ap_pstrndup(data->request->pool, RSTRING(name)->ptr, RSTRING(name)->len);
+    s = apr_pstrndup(data->request->pool, RSTRING(name)->ptr, RSTRING(name)->len);
     ap_soft_timeout(s, data->request);
     return Qnil;
 }
@@ -1299,7 +1350,7 @@ static VALUE request_set_user(VALUE self, VALUE val)
 
     Check_Type(val, T_STRING);
     data = get_request_data(self);
-    data->request->user = ap_pstrndup(data->request->pool,
+    data->request->user = apr_pstrndup(data->request->pool,
 				      RSTRING(val)->ptr,
 				      RSTRING(val)->len);
     return val;
@@ -1393,7 +1444,7 @@ static VALUE request_set_auth_type(VALUE self, VALUE val)
     data = get_request_data(self);
     conf = (core_dir_config *)
 	ap_get_module_config(data->request->per_dir_config, &core_module);
-    conf->ap_auth_type = ap_pstrndup(data->request->pool,
+    conf->ap_auth_type = apr_pstrndup(data->request->pool,
 				     RSTRING(val)->ptr,
 				     RSTRING(val)->len);
     ap_set_module_config(data->request->per_dir_config, &core_module, conf);
@@ -1425,7 +1476,7 @@ static VALUE request_set_auth_name(VALUE self, VALUE val)
     data = get_request_data(self);
     conf = (core_dir_config *)
 	ap_get_module_config(data->request->per_dir_config, &core_module);
-    conf->ap_auth_name = ap_pstrndup(data->request->pool,
+    conf->ap_auth_name = apr_pstrndup(data->request->pool,
 				     RSTRING(val)->ptr,
 				     RSTRING(val)->len);
     ap_set_module_config(data->request->per_dir_config, &core_module, conf);
@@ -1463,23 +1514,44 @@ static VALUE request_bytes_sent(VALUE self)
 static VALUE request_send_fd(VALUE self, VALUE io)
 {
     OpenFile *fptr;
+    request_data *data;
 #ifdef APACHE2
     apr_size_t bytes_sent;
+    apr_file_t *file;
+    int fd;
+    struct stat st;
 #else
+    FILE *f;
     long bytes_sent;
 #endif
-    request_data *data;
 
     request_set_sync(self, Qtrue);
     rb_apache_request_flush(self);    
     data = get_request_data(self);
 
+    Check_Type(io, T_FILE);
     GetOpenFile(io, fptr);
 
 #ifdef APACHE2
-    ap_send_fd((apr_file_t *)fptr->f, data->request, 0, -1, &bytes_sent);
+#if RUBY_VERSION_CODE >= 190
+    fd = fptr->fd;
 #else
-    bytes_sent = ap_send_fd_length(fptr->f, data->request, -1);
+    fd = fileno(fptr->f);
+#endif
+    if (apr_os_file_put(&file, &fd, 0, data->request->pool) != APR_SUCCESS) {
+	rb_raise(rb_eIOError, "apr_os_file_put() failed");
+    }
+    if (fstat(fd, &st) == -1) {
+	rb_sys_fail(fptr->path);
+    }
+    ap_send_fd(file, data->request, 0, st.st_size, &bytes_sent);
+#else
+#if RUBY_VERSION_CODE >= 190
+    f = rb_io_stdio_file(fptr);
+#else
+    f = fptr->f;
+#endif
+    bytes_sent = ap_send_fd_length(f, data->request, -1);
 #endif
 
     return INT2NUM(bytes_sent);
@@ -1536,7 +1608,7 @@ static VALUE request_get_cache_resp(VALUE self)
         data->headers_out = rb_apache_table_new(data->request->headers_out);
 
     Data_Get_Struct(data->headers_out, table, tbl);
-    hdrs_arr = ap_table_elts(tbl);
+    hdrs_arr = apr_table_elts(tbl);
     hdrs = (table_entry *) hdrs_arr->elts;
     for (i = 0; i < hdrs_arr->nelts; i++) {
         if (hdrs[i].key == NULL)
@@ -1561,12 +1633,12 @@ static VALUE request_set_cache_resp(VALUE self, VALUE arg)
 
     Data_Get_Struct(data->headers_out, table, tbl);
     if (arg == Qtrue) {
-        ap_table_setn(tbl, "Pragma", "no-cache");
-	ap_table_setn(tbl, "Cache-control", "no-cache");
+        apr_table_setn(tbl, "Pragma", "no-cache");
+	apr_table_setn(tbl, "Cache-control", "no-cache");
 	return Qtrue;
     } else {
-        ap_table_unset(tbl, "Pragma");
-	ap_table_unset(tbl, "Cache-control");
+        apr_table_unset(tbl, "Pragma");
+	apr_table_unset(tbl, "Cache-control");
 	return Qfalse;
     }
 }
@@ -1615,10 +1687,15 @@ static VALUE call_proc_0(VALUE proc)
 static void call_proc(pool *p, VALUE proc)
 {
 #if APR_HAS_THREADS
-    ruby_call_interpreter(p, (ruby_interp_func_t) call_proc_0,
-			  (void *) proc, NULL, 0);
-#else
-    call_proc_0(proc);
+    if (ruby_is_threaded_mpm) {
+	ruby_call_interpreter(p, (ruby_interp_func_t) call_proc_0,
+			      (void *) proc, NULL, 0);
+    }
+    else {
+#endif
+	call_proc_0(proc);
+#if APR_HAS_THREADS
+    }
 #endif
 }
 
@@ -1651,25 +1728,30 @@ static VALUE request_register_cleanup(int argc, VALUE *argv, VALUE self)
     if (argc == 0)
 	plain_cleanup = rb_f_lambda();
     if (NIL_P(plain_cleanup))
-	plain_cleanup_func = ap_null_cleanup;
+	plain_cleanup_func = apr_pool_cleanup_null;
     else
 	plain_cleanup_func = call_plain_cleanup;
     if (NIL_P(child_cleanup))
-	child_cleanup_func = ap_null_cleanup;
+	child_cleanup_func = apr_pool_cleanup_null;
     else
 	child_cleanup_func = call_child_cleanup;
-    cleanup = ap_palloc(data->request->pool, sizeof(request_cleanup_t));
+    cleanup = apr_palloc(data->request->pool, sizeof(request_cleanup_t));
     cleanup->pool = data->request->pool;
     cleanup->plain_cleanup = plain_cleanup;
     cleanup->child_cleanup = child_cleanup;
-    ap_register_cleanup(data->request->pool, (void *) cleanup,
+    apr_pool_cleanup_register(data->request->pool, (void *) cleanup,
 			plain_cleanup_func, child_cleanup_func);
     
     return Qnil;
 }
 
+static VALUE request_options(VALUE self)
+{
+    request_data *data;
 
-#ifdef HAVE_LIBAPREQ
+    data = get_request_data(self);
+    return data->options;
+}
 
 static VALUE request_libapreq_p( VALUE klass )
 {
@@ -1880,24 +1962,49 @@ static VALUE request_script_path(VALUE self)
 
 static VALUE request_param(VALUE self, VALUE key)
 {
-    request_data *data = get_request_data( self );
+    request_data *data = get_request_data(self);
     const char *val;
 
-    if ( !data->apreq->parsed ) rb_funcall( self, rb_intern("parse"), 0 );
+    if (!data->apreq->parsed) rb_funcall(self, rb_intern("parse"), 0);
     val = ApacheRequest_param(data->apreq, StringValuePtr(key));
     return val ? rb_tainted_str_new2(val) : Qnil;
 }
 
-static VALUE request_params( VALUE self, VALUE key )
+static VALUE request_params(VALUE self, VALUE key)
 {
-    request_data *data = get_request_data( self );
+    request_data *data = get_request_data(self);
     array_header *val;
 
-    if ( !data->apreq->parsed ) rb_funcall( self, rb_intern("parse"), 0 );
+    if (!data->apreq->parsed) rb_funcall(self, rb_intern("parse"), 0);
     val = ApacheRequest_params(data->apreq, StringValuePtr(key));
     return val ? rb_apache_array_new(val) : Qnil;
 }
 
+static int make_all_params(void *data, const char *key, const char *val)
+{
+    VALUE hash = (VALUE) data;
+    VALUE vkey, ary;
+
+    vkey = rb_tainted_str_new2(key);
+    ary = rb_hash_aref(hash, vkey);
+    if (NIL_P(ary)) {
+	ary = rb_ary_new();
+	rb_hash_aset(hash, vkey, ary);
+    }
+    rb_ary_push(ary, rb_tainted_str_new2(val));
+    return 1;
+}
+
+static VALUE request_all_params(VALUE self, VALUE key)
+{
+    request_data *data = get_request_data(self);
+    VALUE result;
+
+    if (!data->apreq->parsed) rb_funcall(self, rb_intern("parse"), 0);
+    result = rb_hash_new();
+    apr_table_do(make_all_params, (void *) result, data->apreq->parms, NULL);
+    return result;
+}
 
 static VALUE request_paramtable( VALUE self )
 {
@@ -1953,16 +2060,6 @@ static VALUE request_cookies_eq( VALUE self, VALUE newhash )
     return newhash;
 }
 
-#else /* !HAVE_LIBAPREQ */
-
-static VALUE request_libapreq_p( VALUE klass )
-{
-    return Qfalse;
-}
-
-#endif /* HAVE_LIBAPREQ */
-
-
 void rb_init_apache_request()
 {
     rb_eApachePrematureChunkEndError =
@@ -2150,10 +2247,10 @@ void rb_init_apache_request()
     rb_define_method(rb_cApacheRequest, "lookup_file", request_lookup_file, 1);
     rb_define_method(rb_cApacheRequest, "register_cleanup",
 		     request_register_cleanup, -1);
+    rb_define_method(rb_cApacheRequest, "options", request_options, 0);
 
     rb_define_singleton_method(rb_cApacheRequest, "libapreq?", request_libapreq_p, 0 );
 
-#ifdef HAVE_LIBAPREQ
     id_post_max = rb_intern( "post_max" );
     id_disable_uploads = rb_intern( "disable_uploads" );
     id_temp_dir = rb_intern( "temp_dir" );
@@ -2165,6 +2262,7 @@ void rb_init_apache_request()
     rb_define_method(rb_cApacheRequest, "script_path", request_script_path, 0);
     rb_define_method(rb_cApacheRequest, "param", request_param, 1);
     rb_define_method(rb_cApacheRequest, "params", request_params, 1);
+    rb_define_method(rb_cApacheRequest, "all_params", request_all_params, 0);
     rb_define_method(rb_cApacheRequest, "paramtable", request_paramtable, 0);
     rb_define_method(rb_cApacheRequest, "params_as_string",
 		     request_params_as_string, 1);
@@ -2183,7 +2281,6 @@ void rb_init_apache_request()
     rb_define_method(rb_cApacheRequest, "upload_hook", request_upload_hook, 0);
     rb_define_method(rb_cApacheRequest, "upload_hook_data=", request_upload_hook_data_eq, 1);
     rb_define_method(rb_cApacheRequest, "upload_hook_data", request_upload_hook_data, 0);
-#endif
 }
 
 /*
@@ -2192,3 +2289,5 @@ void rb_init_apache_request()
  * tab-width: 8
  * End:
  */
+
+/* vim: set filetype=c ts=8 sw=4 : */
